﻿using System;
using System.Collections.Generic;
using GrapeCity.Competition.Maze.Engine.Explorers;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Engine.Elements
{
    #region The Trigger class
    /// <summary>
    ///   Represents the element indicate a trigger.
    /// </summary>
    [Serializable]
    public class Trigger : MazeElement
    {
        /// <summary>
        ///   Indicates the doors operated by the trigger.
        /// </summary>
        private readonly List<Position> m_doors = new List<Position>();
        private TriggerStyle m_style = TriggerStyle.OpenOrClose;

        #region Constructor
        /// <summary>
        ///   Initialize the <see cref="Trigger"/> object.
        /// </summary>
        /// <param name="mazeMap">
        ///   A <see cref="IMazeMap"/> object indicates the maze map.
        /// </param>
        public Trigger(IMazeMap mazeMap)
            : base(mazeMap)
        {
        }
        #endregion

        #region Clone
        /// <summary>
        ///   Creates a new object that is a copy of the current instance.
        /// </summary>
        /// <returns>
        ///   A new object that is a copy of this instance.
        /// </returns>
        public override MazeElement Clone(IMazeMap map)
        {
            Trigger trigger = new Trigger(map);
            trigger.m_doors.AddRange(this.m_doors);
            trigger.m_style = this.m_style;
            return trigger;
        }
        #endregion

        #region Properties
        /// <summary>
        ///   Gets the doors operated by the trigger.
        /// </summary>
        /// <value>
        ///   A <see cref="Position"/> collection contains all doors operated by the trigger.
        /// </value>
        public List<Position> Doors
        {
            get
            {
                return this.m_doors;
            }
        }

        /// <summary>
        ///   Gets the information in the position.
        /// </summary>
        /// <value>
        ///   A <see cref="PositionInfo"/> value indicates the position information.
        /// </value>
        public override PositionInfo ElementInfo
        {
            get
            {
                return PositionInfo.Trigger;
            }
        }


        /// <summary>
        /// Gets or sets the trigger style.
        /// </summary>
        /// <value>
        ///     A <see cref="TriggerStyle"/> indicates the trigger style.
        /// </value>
        public TriggerStyle Style
        {
            get { return m_style; }
            set { m_style = value; }
        }

        #endregion

        #region Methods
        /// <summary>
        ///   Link the door.
        /// </summary>
        /// <param name="doorPosition">
        ///   An <see cref="IPosition"/> object indicates the door position.
        /// </param>
        public void Link(IPosition doorPosition)
        {
            Position position = new Position(doorPosition);
            if (!this.Doors.Contains(position))
                this.Doors.Add(position);
        }
        /// <summary>
        ///   Unlink the door.
        /// </summary>
        /// <param name="doorPosition">
        ///   An <see cref="IPosition"/> object indicates the door position.
        /// </param>
        public void UnLink(IPosition doorPosition)
        {
            Position position = new Position(doorPosition);
            if (this.Doors.Contains(position))
                this.Doors.Remove(position);
        }

        /// <summary>
        ///   Occurs when an explorer try to enter the position.
        /// </summary>
        /// <param name="explorer">
        ///   An <see cref="IExplorer"/> object indicates the explorer.
        /// </param>
        /// <returns>
        ///   <b>true</b>, if the explorer can enter the position; otherwise, <b>false</b>.
        /// </returns>
        public override bool CanExplorerIn(IMazeExplorer explorer)
        {
            return true;
        }

        /// <summary>
        ///   Occurs when a explorer stay at the position.
        /// </summary>
        /// <param name="explorer">
        ///   A <see cref="IExplorer"/> object indicates the explorer.
        /// </param> 
        public override void OnExplorerStay(IMazeExplorer explorer)
        {
            foreach (Position door in this.Doors)
            {
                Door element = this.MazeMap.GetElement(door) as Door;
                if (element != null)
                {
                    if (Style == TriggerStyle.OpenOrClose)
                    {
                        if (element.IsLocked)
                            element.Open();
                        else
                            element.Close();
                    }
                    else if (Style == TriggerStyle.AlwaysClose)
                    {
                        element.Close();
                    }
                    else
                    {
                        element.Open();
                    }
                }
            }
        }
        #endregion
    }
    #endregion
}