﻿using System.Diagnostics;

namespace GrapeCity.Competition.Maze.Engine.Explorers
{
    #region The MazeExplorer class

    /// <summary>
    ///   Represents the explorer.
    /// </summary>
    public class MazeExplorer : IMazeExplorer
    {
        #region Instance Data

        /// <summary>
        ///   Indicates the explorer name.
        /// </summary>
        private readonly string m_name = string.Empty;

        private IMazeEngine m_game;
        private Position m_startPosition;
        private string m_thinker;
        private Position m_position;
        private int m_steps;
        private int m_looks;
        private int m_blocks;

        #endregion

        #region Properties

        public bool InGame
        {
            get { return this.m_game != null; }
        }

        public string Thinker
        {
            get { return m_thinker; }
            set { m_thinker = value; }
        }

        public int Looks
        {
            get { return this.m_looks; }
        }

        public int Blocks
        {
            get { return this.m_blocks; }
        }

        public int X
        {
            get { return m_position.X; }
            set { m_position.X = value; }
        }

        public int Y
        {
            get { return m_position.Y; }
            set { m_position.Y = value; }
        }

        #endregion

        #region Constructor

        public MazeExplorer(string name, string thinkerName)
        {
            Trace.Assert(!string.IsNullOrEmpty(name));

            this.m_name = name;
            this.m_steps = 0;
            this.m_looks = 0;
            this.m_blocks = 0;
            this.m_thinker = thinkerName;
        }

        #endregion

        #region IExplorer Members

        public string Name
        {
            get { return m_name; }
        }

        public int Steps
        {
            get { return m_steps; }
        }

        public PositionInfo Current
        {
            get { return Look(Direction.None); }
        }

        public IPosition Location
        {
            get { return m_position; }
        }

        public PositionInfo Look(Direction direction)
        {
            System.Threading.Interlocked.Increment(ref m_looks);
            return m_game.Look(this, direction);
        }

        public void Say(string message, object information)
        {
            m_game.Say(this, message, information);
        }

        public bool Whisper(string name, string message, object information)
        {
            return m_game.Wisper(this, name, message, information);
        }

        public void Stay()
        {
            Move(Direction.None);
        }

        public bool Move(Direction direction)
        {
            if (m_game.Move(this, direction))
            {
                System.Threading.Interlocked.Increment(ref m_steps);
                return true;
            }

            System.Threading.Interlocked.Increment(ref m_blocks);
            return false;
        }

        #endregion

        #region Methods

        public void JoinGame(IPosition position, IMazeEngine maze)
        {
            this.m_game = maze;
            this.m_startPosition = new Position(position);
            this.m_position = new Position(position);
        }

        public void LeaveGame()
        {
            this.m_game = null;
            this.m_startPosition = Position.Zero;
            this.m_position = Position.Zero;
        }

        public void Reset()
        {
            m_steps = 0;
            m_looks = 0;
            m_blocks = 0;
            m_position = new Position(m_startPosition);
        }

        public virtual bool CanMove()
        {
            return true;
        }

        #endregion
    } 
    #endregion
}