﻿using System;

namespace GrapeCity.Competition.Maze.Engine.Explorers
{
    #region The MazeExplorerClient class
    internal sealed class MazeExplorerClient : IExplorer
    {
        private readonly WeakReference explorer;

        #region Explorer
        private IExplorer Explorer
        {
            get
            {
                if (!explorer.IsAlive)
                    throw new ObjectDisposedException("Explorer");

                IExplorer e = explorer.Target as IExplorer;
                if (e == null)
                    throw new NullReferenceException("Explorer");

                return e;
            }
        }
        #endregion

        internal MazeExplorerClient(IExplorer explorer)
        {
            this.explorer = new WeakReference(explorer);
        }

        #region IExplorer Members

        /// <summary>
        ///   Gets the explorer name.
        /// </summary>
        /// <value>
        ///   A <b>string</b> value indicates the explorer name.
        /// </value>
        public string Name
        {
            get { return Explorer.Name; }
        }

        /// <summary>
        ///   Gets the total steps
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the total steps.
        /// </value>
        public int Steps
        {
            get { return Explorer.Steps; }
        }

        /// <summary>
        ///   Gets the information of current position.
        /// </summary>
        /// <value>
        ///   A <see cref="PositionInfo"/> value indicates the information of current position.
        /// </value>
        public PositionInfo Current
        {
            get { return Explorer.Current; }
        }

        public IPosition Location
        {
            get { return Explorer.Location; }
        }

        public int X
        {
            get { return Explorer.Location.X; }
        }

        public int Y
        {
            get { return Explorer.Location.Y; }
        }

        /// <summary>
        ///   Gets the information of surrounding position. 
        /// </summary>
        /// <param name="direction">
        ///   A <see cref="Direction"/> value indicates the surrounding position. 
        /// </param>
        /// <returns>
        ///   A <see cref="PositionInfo"/> value indicates the information of surrounding position.
        /// </returns>
        public PositionInfo Look(Direction direction)
        {
            return Explorer.Look(direction);
        }

        /// <summary>
        ///   Say something to other explorer in the maze.
        /// </summary>
        /// <param name="message">
        ///   A <b>string</b> value indicates the message.
        /// </param>
        /// <param name="information">
        ///   A <see cref="System.Object"/> indicates the extra information.
        /// </param>
        public void Say(string message, object information)
        {
            Explorer.Say(message, information);
        }

        /// <summary>
        ///   Whisper something to other explorer in the maze.
        /// </summary>
        /// <param name="name">
        ///   A <b>string</b> value indicates the listener's name.
        /// </param>
        /// <param name="message">
        ///   A <b>string</b> value indicates the message.
        /// </param>
        /// <param name="information">
        ///   A <see cref="System.Object"/> indicates the extra information.
        /// </param>
        public bool Whisper(string name, string message, object information)
        {
            return Explorer.Whisper(name, message, information);
        }

        /// <summary>
        ///   Stay in current position. 
        /// </summary>
        public void Stay()
        {
            Explorer.Stay();
        }

        /// <summary>
        ///   Move to next position.
        /// </summary>
        /// <param name="direction">
        ///   A <see cref="Direction"/> value indicates the moving direction.
        /// </param>
        /// <returns>
        ///   <b>true</b>, if the moving is successful; otherwise, <b>false</b>.
        /// </returns>
        public bool Move(Direction direction)
        {
            return Explorer.Move(direction);
        }

        #endregion
    } 
    #endregion
}