﻿using System;
using System.Diagnostics;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Engine.Maps
{
    #region The Map class

    /// <summary>
    ///   Represents the maze map.
    /// </summary>
    [Serializable]
    public class MazeMap : IMazeMapEditor, ICloneable
    {
        #region Instance Data

        /// <summary>
        ///   Indicates the map name.
        /// </summary>
        private readonly string m_name = string.Empty;

        /// <summary>
        ///   Indicates the entrance position.
        /// </summary>
        private readonly IPosition m_entrance = Position.Zero;

        /// <summary>
        ///   Indicates the exit position.
        /// </summary>
        private readonly IPosition m_exit = Position.Zero;

        /// <summary>
        ///   Indicates the storage kept the all element.
        /// </summary>
        private readonly MazeElement[,] m_mazeElements = null;

        #endregion

        #region Constructor

        /// <summary>
        ///   Initialize the maze map.
        /// </summary>
        /// <param name="name">
        ///   A <b>string</b> value indicates the map name.
        /// </param>
        public MazeMap(string name)
            : this(name, 10, 10)
        {
        }

        /// <summary>
        /// Initialize the maze map.
        /// </summary>
        /// <param name="name">
        /// A <b>string</b> value indicates the map name.
        /// </param>
        /// <param name="width">
        /// An <b>int</b> value indicates the maze width.
        /// </param>
        /// <param name="height">
        /// An <b>int</b> value indicates the maze height.
        /// </param>
        public MazeMap(string name, int width, int height)
            : this(name, width, height, new Position(width - 1, height - 1), new Position(0, 0))
        {
        }

        /// <summary>
        ///   Initialize the maze map.
        /// </summary>
        /// <param name="name">
        ///   A <b>string</b> value indicates the map name.
        /// </param>
        /// <param name="width">
        ///   An <b>int</b> value indicates the maze width.
        /// </param>
        /// <param name="height">
        ///   An <b>int</b> value indicates the maze height.
        /// </param>
        /// <param name="entrance">
        ///   An <see cref="IPosition"/> value indicates the entrance position.
        /// </param>
        /// <param name="exit">
        ///   An <see cref="IPosition"/> value indicates the exit position.
        /// </param>
        public MazeMap(string name, int width, int height, IPosition entrance, IPosition exit)
        {
            Trace.Assert((name != null) && (name.Length != 0));
            Trace.Assert(width >= 5);
            Trace.Assert(height >= 5);

            this.m_name = name;
            this.m_entrance = entrance;
            this.m_exit = exit;
            this.m_mazeElements = new MazeElement[width,height];

            bool entranceCreated = false;
            bool exitCreated = false;

            for(int x=0;x<width;x++)
                for(int y=0;y<height;y++)
                {
                    if (x == m_entrance.X && y == m_entrance.Y)
                    {
                        this.m_mazeElements[x, y] = new Entrance(this);
                        entranceCreated = true;
                    }
                    else if (x == m_exit.X && y == m_exit.Y)
                    {
                        this.m_mazeElements[x, y] = new Exit(this);
                        exitCreated = true;
                    }
                    else
                    {
                        this.m_mazeElements[x, y] = new Path(this);
                    }
                }

            Trace.Assert(entranceCreated);
            Trace.Assert(exitCreated);
        }

        #endregion

        #region Properties

        /// <summary>
        ///   Gets the map name.
        /// </summary>
        /// <value>
        ///   A <b>string</b> value indicates the map name.
        /// </value>
        public string Name
        {
            get { return this.m_name; }
        }

        /// <summary>
        ///   Gets the width of maze.
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the maze width.
        /// </value>
        public int Width
        {
            get { return this.m_mazeElements.GetLength(0); }
        }

        /// <summary>
        ///   Gets the height of maze.
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the maze height.
        /// </value>
        public int Height
        {
            get { return this.m_mazeElements.GetLength(1); }
        }

        /// <summary>
        ///   Gets the entrance position of maze.
        /// </summary>
        /// <value>
        ///   An <see cref="IPosition"/> value indicates the entrance position of maze.
        /// </value>
        public IPosition Entrance
        {
            get { return this.m_entrance; }
        }

        /// <summary>
        ///   Gets the exit position of maze.
        /// </summary>
        /// <value>
        ///   An <see cref="IPosition"/> value indicates the exit position of maze.
        /// </value>
        public IPosition Exit
        {
            get { return this.m_exit; }
        }

        #endregion

        #region Methods

        /// <summary>
        ///   Check whether it is a valid position.
        /// </summary>
        /// <param name="position">
        ///   A <see cref="IPosition"/> indicates the specified position.
        /// </param>
        /// <returns>
        ///   <b>true</b>, if it is a valid position; otherwise, <b>false</b>.
        /// </returns>
        public bool IsValidPosition(IPosition position)
        {
            return this.IsValidPosition(position.X, position.Y);
        }

        /// <summary>
        ///   Check whether it is a valid position.
        /// </summary>
        /// <param name="x">
        ///   The x-coordinate of the upper-left corner of the element to check. 
        /// </param>
        /// <param name="y">
        ///  The y-coordinate of the upper-left corner of the element to check. 
        /// </param>
        /// <returns>
        ///   <b>true</b>, if it is a valid position; otherwise, <b>false</b>.
        /// </returns>
        public bool IsValidPosition(int x, int y)
        {
            if ((x < 0) || (x >= this.Width))
                return false;
            else if ((y < 0) || (y >= this.Height))
                return false;
            else
                return true;
        }

        /// <summary>
        ///   Gets the element based specified position.
        /// </summary>
        /// <param name="position">
        ///   A <see cref="IPosition"/> value indicates the specified position.
        /// </param>
        /// <returns>
        ///   A <see cref="MazeElement"/> object indicates the element.
        /// </returns>
        public MazeElement GetElement(IPosition position)
        {
            return this.GetElement(position.X, position.Y);
        }

        /// <summary>
        ///   Gets the element specified by a coordinate pair.
        /// </summary>
        /// <param name="x">
        ///   The x-coordinate of the upper-left corner of the element to get. 
        /// </param>
        /// <param name="y">
        ///  The y-coordinate of the upper-left corner of the element to get. 
        /// </param>
        /// <returns>
        ///   A <see cref="MazeElement"/> object indicates the element.
        /// </returns>
        public MazeElement GetElement(int x, int y)
        {
            if (this.IsValidPosition(x, y))
                return this.m_mazeElements[x, y];
            else
                return new Wall(this);
        }

        /// <summary>
        ///   Set the element on specified position.
        /// </summary>
        /// <param name="position">
        ///   A <see cref="IPosition"/> value indicates the specified position.
        /// </param>
        /// <param name="element">
        ///   A <see cref="MazeElement"/> object indicates the element.
        /// </param>
        public void SetElement(IPosition position, MazeElement element)
        {
            SetElement(position.X, position.Y, element);
        }

        /// <summary>
        ///   Sets the element specified by a coordinate pair.
        /// </summary>
        /// <param name="x">
        ///   The x-coordinate of the upper-left corner of the element to set. 
        /// </param>
        /// <param name="y">
        ///  The y-coordinate of the upper-left corner of the element to set. 
        /// </param>
        /// <param name="element">
        ///  A <see cref="MazeElement"/> object indicates the element.
        /// </param>
        public void SetElement(int x, int y, MazeElement element)
        {
            Trace.Assert(this.IsValidPosition(x, y));
            Trace.Assert((this.Entrance.X != x) || (this.Entrance.Y != y));
            Trace.Assert((this.Exit.X != x) || (this.Exit.Y != y));

            this.m_mazeElements[x, y] = element;
        }

        #endregion

        #region ICloneable
        public object Clone()
        {
            MazeMap map = new MazeMap(this.Name, this.Width, this.Height, this.Entrance, this.Exit);
            int height, width;
            width = this.Width;
            height = this.Height;

            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    if (this.m_mazeElements[x, y] != null)
                    {
                        map.m_mazeElements[x, y] = this.m_mazeElements[x, y].Clone(map);
                    }
                }
            }

            return map;
        } 
        #endregion
    }

    #endregion
}
