﻿using System;
using GrapeCity.Competition.Maze.Engine.Explorers;
using GrapeCity.Competition.Maze.Engine.Maps;
using System.Diagnostics;

namespace GrapeCity.Competition.Maze.Engine
{
    #region The MazeElement class

    /// <summary>
    ///   Represents a element in maze.
    /// </summary>
    [Serializable]
    public abstract class MazeElement : IPosition
    {
        #region IPosition Members

        /// <summary>
        ///   Gets the x value in coordinate system.
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the x value in coordinate system.
        /// </value>
        int IPosition.X
        {
            get { return this.X; }
        }

        /// <summary>
        ///   Gets the y value in coordinate system.
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the y value in coordinate system.
        /// </value>
        int IPosition.Y
        {
            get { return this.Y; }
        }

        #endregion

        #region Instance Data

        /// <summary>
        ///   Indicates the maze map.
        /// </summary>
        private IMazeMap m_mazeMap = null;

        #endregion

        #region Constructor

        /// <summary>
        ///   Initialize the <see cref="MazeElement"/> object.
        /// </summary>
        /// <param name="mazeMap">
        ///   An <see cref="IMazeMap"/> object indicates the maze map.
        /// </param>
        protected MazeElement(IMazeMap mazeMap)
        {
            Trace.Assert(mazeMap != null);

            this.m_mazeMap = mazeMap;
        }

        #endregion

        #region Clone

        /// <summary>
        ///   Creates a new object that is a copy of the current instance.
        /// </summary>
        /// <returns>
        ///   A new object that is a copy of this instance.
        /// </returns>
        public abstract MazeElement Clone(IMazeMap map);

        #endregion

        #region Properties

        /// <summary>
        ///   Gets the maze map.
        /// </summary>
        /// <value>
        ///   An <see cref="IMazeMap"/> object indicates the maze map.
        /// </value>
        protected IMazeMap MazeMap
        {
            get { return this.m_mazeMap; }
        }

        /// <summary>
        ///   Gets the x value in coordinate system.
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the x value in coordinate system.
        /// </value>
        protected int X
        {
            get
            {
                for (int i = 0; i < this.MazeMap.Width; i++)
                    for (int j = 0; j < this.MazeMap.Height; j++)
                    {
                        Position position = new Position(i, j);
                        if (this.MazeMap.GetElement(position) == this)
                            return position.X;
                    }

                return -1;
            }
        }

        /// <summary>
        ///   Gets the y value in coordinate system.
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the y value in coordinate system.
        /// </value>
        protected int Y
        {
            get
            {
                for (int i = 0; i < this.MazeMap.Width; i++)
                    for (int j = 0; j < this.MazeMap.Height; j++)
                    {
                        Position position = new Position(i, j);
                        if (this.MazeMap.GetElement(position) == this)
                            return position.Y;
                    }

                return -1;
            }
        }

        /// <summary>
        ///   Gets the information in the position.
        /// </summary>
        /// <value>
        ///   A <see cref="PositionInfo"/> value indicates the position information.
        /// </value>
        public abstract PositionInfo ElementInfo { get; }

        #endregion

        #region Methods

        /// <summary>
        ///   Occurs when an explorer try to enter the position.
        /// </summary>
        /// <param name="explorer">
        ///   An <see cref="IExplorer"/> object indicates the explorer.
        /// </param>
        /// <returns>
        ///   <b>true</b>, if the explorer can enter the position; otherwise, <b>false</b>.
        /// </returns>
        public virtual bool CanExplorerIn(IMazeExplorer explorer)
        {
            return false;
        }

        /// <summary>
        ///   Occurs when an explorer try to leave the position.
        /// </summary>
        /// <param name="explorer">
        ///   An <see cref="IExplorer"/> object indicates the explorer.
        /// </param>
        /// <returns>
        ///   <b>true</b>, if the explorer can leave the position; otherwise, <b>false</b>.
        /// </returns>
        public virtual bool CanExplorerOut(IMazeExplorer explorer)
        {
            return true;
        }

        /// <summary>
        ///   Occurs when an explorer enter the position.
        /// </summary>
        /// <param name="explorer">
        ///   An <see cref="IExplorer"/> object indicates the explorer.
        /// </param>
        public virtual void OnExplorerIn(IMazeExplorer explorer)
        {
        }

        /// <summary>
        ///   Occurs when an explorer leave the position.
        /// </summary>
        /// <param name="explorer">
        ///   An <see cref="IExplorer"/> object indicates the explorer.
        /// </param>
        public virtual void OnExplorerOut(IMazeExplorer explorer)
        {
        }

        /// <summary>
        ///   Occurs when an explorer stay at the position.
        /// </summary>
        /// <param name="explorer">
        ///   An <see cref="IExplorer"/> object indicates the explorer.
        /// </param> 
        public virtual void OnExplorerStay(IMazeExplorer explorer)
        {
        }

        #endregion
    }

    #endregion
}
