﻿using System;
using System.Diagnostics;
using GrapeCity.Competition.Maze.Engine.Explorers;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Engine
{
    #region The MazeEngine Class
    public class MazeEngine : Maze2D, IMazeEngine
    {
        public MazeEngine(MazeMap maze)
            : base(maze)
        {
        }

        #region Create/Remove Explorer

        public MazeExplorer CreateExplorer(Type thinkerType)
        {
            string namePrefix = NameAttribute.GetName(thinkerType);
            int nameSuffix = 1;
            string explorerName;
            do
            {
                explorerName = string.Format("{0}({1})", namePrefix, nameSuffix);
                nameSuffix++;
            } while (Explorers.ContainsKey(explorerName));

            return CreateExplorer(explorerName, thinkerType);
        }

        public MazeExplorer CreateExplorer(string name, Type thinkerType)
        {
            Trace.Assert(!string.IsNullOrEmpty(name));

            if (thinkerType == null)
                throw new NullReferenceException("thinker type is null.");

            if (Explorers.ContainsKey(name))
                throw new InvalidOperationException(string.Format("The explorer '{0}' is already exists.", name));

            //string thinkerName = thinkerType.AssemblyQualifiedName;
            string thinkerName = name;
            try
            {
                MazeExplorer mazeExplorer = new MazeExplorer(name, thinkerName);
                Explorers.Add(name, mazeExplorer);

                if (!Thinkers.ContainsKey(thinkerName))
                    Thinkers.Add(thinkerName, new Thinker(thinkerType));

                return mazeExplorer;
            }
            catch (Exception e)
            {
                RemoveExplorer(name);
                throw new TypeInitializationException(thinkerName, e);
            }
        }

        public void RemoveExplorer(string name)
        {
            if (Explorers.ContainsKey(name))
            {
                MazeExplorer mazeExplorer = Explorers[name];

                if (mazeExplorer.InGame)
                {
                    mazeExplorer.LeaveGame();
                }

                Explorers.Remove(name);
            }
        }

        #endregion

        #region Explorer says

        void IMazeEngine.Say(MazeExplorer mazeExplorer, string message, object information)
        {
            OnExplorerSay(mazeExplorer, message, information);
        }

        protected virtual void OnExplorerSay(MazeExplorer e, string message, object information)
        {
            MazeExplorer[] list = new MazeExplorer[Explorers.Count];
            Explorers.Values.CopyTo(list, 0);

            //calc distance
            foreach (MazeExplorer explorer in list)
            {
                if (explorer.Name == e.Name) continue;

                try
                {
                    Thinker thinker;
                    if (Thinkers.TryGetValue(explorer.Thinker, out thinker))
                    {
                        thinker.Hear(e.Name, message, information);
                    }
                }
                finally
                {
                }
            }
        }

        #endregion
        #region Explorer wisper

        bool IMazeEngine.Wisper(MazeExplorer mazeExplorer, string name, string message, object information)
        {
            return OnExplorerWisper(mazeExplorer, name, message, information);
        }

        protected virtual bool OnExplorerWisper(MazeExplorer e, string name, string message, object information)
        {
            MazeExplorer mazeExplorer;
            if (!Explorers.TryGetValue(name, out mazeExplorer) || mazeExplorer == null)
                return false;

            Thinker thinker;
            if (!Thinkers.TryGetValue(name, out thinker) || thinker == null)
                return false;

            try
            {
                thinker.Hear(e.Name, message, information);
                return true;
            }
            catch
            {
                return false;
            }
        }

        #endregion
        #region Explorer looks

        PositionInfo IMazeEngine.Look(MazeExplorer mazeExplorer, Direction direction)
        {
            return OnExplorerLook(mazeExplorer, direction);
        }

        protected virtual PositionInfo OnExplorerLook(MazeExplorer e, Direction direction)
        {
            Position lookingPosition = Position.GetPositionByDirection(e, direction);

            if (Map.IsValidPosition(lookingPosition))
            {
                MazeElement element = Map.GetElement(lookingPosition);
                if (element != null)
                    return element.ElementInfo;
            }

            return PositionInfo.Wall;
        }

        #endregion
        #region Explorer moves

        bool IMazeEngine.Move(MazeExplorer mazeExplorer, Direction direction)
        {
            return OnExplorerMove(mazeExplorer, direction);
        }

        protected virtual bool OnExplorerMove(MazeExplorer e, Direction direction)
        {
            Position fromPosition = Position.GetPositionByDirection(e, Direction.None);
            Position toPosition = Position.GetPositionByDirection(e, direction);

            if (Map.IsValidPosition(fromPosition) && Map.IsValidPosition(toPosition))
            {
                MazeElement fromElement = Map.GetElement(fromPosition);
                MazeElement toElement = Map.GetElement(toPosition);

                if (fromElement != null && toElement != null)
                {
                    bool canIn = toElement.CanExplorerIn(e);
                    bool canOut = fromElement.CanExplorerOut(e);
                    bool canMove = e.CanMove();

                    return canIn && canOut && canMove;
                }
            }

            return false;
        }

        #endregion
    } 
    #endregion
}
