﻿using System;
using System.Collections.Generic;
using System.Threading;
using GrapeCity.Competition.Maze.Engine;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Explorers;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Engine
{
    #region The MazeGame class
    /// <summary>
    ///   Represents the maze engine.
    /// </summary>
    public class MazeGame : MazeEngine, IMazeGame
    {
        private long runningCount = 0;
        private int interval;
        private bool running;
        private bool paused;

        private int totalBlocks = 0;
        private readonly Dictionary<int, Position> reachBlocks;

        protected MazeGame(MazeMap maze) : base(maze)
        {
            this.reachBlocks = new Dictionary<int, Position>();
            this.interval = 25;
            this.running = false;
            this.paused = false;
            this.InitialReachs();
        }

        protected void InitialReachs()
        {
            this.totalBlocks = 0;
            this.reachBlocks.Clear();

            int width = Map.Width;
            int height = Map.Height;

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    if (!(Map.GetElement(x, y) is Wall))
                    {
                        totalBlocks++;
                    }
                }
            }
        }

        public int TotalBlocks
        {
            get { return totalBlocks; }
        }

        public int ReachBlocks
        {
            get { return reachBlocks.Count; }
        }

        public long RunningCount
        {
            get { return runningCount; }
        }

        public MazeExplorer[] MazeExplorers
        {
            get
            {
                MazeExplorer[] array = new MazeExplorer[Explorers.Count];
                Explorers.Values.CopyTo(array, 0);
                return array;
            }
        }

        public Thinker[] MazeThinkers
        {
            get
            {
                Thinker[] array = new Thinker[Thinkers.Count];
                Thinkers.Values.CopyTo(array, 0);
                return array;
            }
        }

        #region IMazeGame Members

        public int Interval
        {
            get { return interval; }
            set { interval = value; }
        }

        public bool Running
        {
            get { return running; }
        }

        public bool Paused
        {
            get { return paused; }
            set { paused = value; }
        }

        public void Start()
        {
            if (running || runningCount != 0) return;
            running = true;
            paused = false;

            InitialReachs();
            OnGameStart();
            foreach (MazeExplorer e in Explorers.Values)
            {
                e.JoinGame(this.Map.Entrance, this);
                ThreadPool.QueueUserWorkItem(this.ExecuteThinker, e.Name);
            }
        }

        protected virtual void OnGameStart()
        {
        }

        public void Stop()
        {
            if (!running) return;

            foreach (MazeExplorer value in Explorers.Values)
            {
                Thinker thinker;
                if (!Thinkers.TryGetValue(value.Thinker, out thinker) || thinker == null)
                    return;

                try
                {
                    thinker.Terminate();
                }
                finally
                {
                }
            }

            running = false;
            paused = false;
            OnGameStop();
        }

        protected virtual void OnGameStop()
        {
            
        }

        public void Reset()
        {
            Stop();

            foreach (MazeExplorer value in Explorers.Values)
                value.Reset();

            reachBlocks.Clear();

            ResetMap();

        }

        #endregion

        #region void ExecuteThinker(object)

        protected void ExecuteThinker(object context)
        {
            string explorerName = context as string;
            if (explorerName != null)
            {
                MazeExplorer e;
                if (!Explorers.TryGetValue(explorerName, out e) || e == null)
                    return;

                Thinker thinker;
                if (!Thinkers.TryGetValue(e.Thinker, out thinker) || thinker == null)
                    return;
 
                try
                {
                    Interlocked.Increment(ref runningCount);
                    thinker.Initialize();
                    thinker.Think(new MazeMapClient(Map), new MazeExplorerClient(e));
                }
                finally
                {
                    Interlocked.Decrement(ref runningCount);
                    thinker.Terminate();
                }
            }
        }

        #endregion

        protected override PositionInfo OnExplorerLook(MazeExplorer e, Direction direction)
        {
            if (!running) return GrapeCity.Competition.Maze.PositionInfo.Path;
            PositionInfo PositionInfo = base.OnExplorerLook(e, direction);
            //OnAction();
            return PositionInfo;
        }

        protected override bool OnExplorerMove(MazeExplorer e, Direction direction)
        {
            if (!running) return false;
            bool result = base.OnExplorerMove(e, direction);
            if(result)
            {
                Position toPosition = Position.GetPositionByDirection(e, direction);
                e.X = toPosition.X;
                e.Y = toPosition.Y;
                
                int hashcode = toPosition.ToString().GetHashCode();

                if (!reachBlocks.ContainsKey(hashcode)){
                    reachBlocks.Add(hashcode, toPosition);
                }

                MazeElement element = Map.GetElement(toPosition);
                if (element != null) element.OnExplorerStay(e);
            }

            OnAction();
            return result;
        }

        private void OnAction()
        {
            if (Paused) while (Paused) Thread.Sleep(25);

            OnUpdate();

            if (Interval > 0)
            {
                //if (nextRenderTick < Time.Milliseconds)
                //{
                //    lastRenderTick = nextRenderTick;
                //    nextRenderTick = Time.Milliseconds + Interval;
                //    OnRender();
                //}


                long startMs = Time.Milliseconds;
                OnRender();
                long spendMs = Math.Max(0, Time.Milliseconds - startMs);
                int sleepMs = (int)(interval - spendMs);

                if (sleepMs > 25) Thread.Sleep(sleepMs);
            }
        }

        protected virtual void OnUpdate()
        {
            
        }

        protected virtual void OnRender()
        {
        }
    }

    #endregion
}
