﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.Engine;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Generators
{
    public partial class ElementEditorDialog : Form
    {
        #region Position Type
        private class PositionType
        {
            private readonly string name;
            private Position position;

            public PositionType(int x, int y, MazeElement e)
            {
                position = new Position(x, y);

                if (e != null)
                {
                    name = position.ToString() + e.GetType().Name;
                }
                else
                {
                    name = position.ToString();
                }
            }

            public Position Position
            {
                get { return position; }
            }

            public override string ToString()
            {
                return name;
            }
        } 
        #endregion

        #region Door Info
        private class DoorInfo
        {
            private Position p;
            private Door d;
            private string tostring;

            public DoorInfo(int x, int y, Door door)
            {
                p = new Position(x, y);
                d = door;
                tostring = p.ToString();
            }

            public Door Door
            {
                get { return d; }
            }

            public Position Position
            {
                get { return p; }
            }

            public override string ToString()
            {
                return tostring;
            }
        } 
        #endregion

        private readonly IMazeMap editingMap;
        private readonly Door editingDoor;
        private readonly Trigger editingTrigger;
        private readonly List<DoorInfo> editingDoors;
        private readonly Transport editingTransport;

        private ElementEditorDialog()
        {
            InitializeComponent();
        }

        #region Edit Door

        public ElementEditorDialog(Door door)
            : this()
        {
            tabControl.TabPages.Clear();
            tabControl.TabPages.Add(doorTabPage);
            editingDoor = door;
            doorStateCheckBox.DataBindings.Add("Checked", editingDoor, "IsLocked");
        }

        #endregion
        #region Edit Trigger

        public ElementEditorDialog(Trigger trigger, IMazeMap map)
            : this()
        {
            tabControl.TabPages.Clear();
            tabControl.TabPages.Add(triggerTabPage);

            editingTrigger = trigger;
            editingMap = map;

            //controls
            triggerStyleComboBox.Items.Clear();
            triggerStyleComboBox.Items.Add(TriggerStyle.AlwaysOpen);
            triggerStyleComboBox.Items.Add(TriggerStyle.AlwaysClose);
            triggerStyleComboBox.Items.Add(TriggerStyle.OpenOrClose);
            triggerStyleComboBox.DataBindings.Add("SelectedItem", trigger, "Style");

            editingDoors = InitializeDoorNodes(map);
            RefreshLinkList();
        }

        private static List<DoorInfo> InitializeDoorNodes(IMazeMap map)
        {
            List<DoorInfo> doors = new List<DoorInfo>();
            int width = map.Width;
            int height = map.Height;

            //draw element
            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    MazeElement e = map.GetElement(x, y);
                    if (e != null && e is Door)
                    {
                        doors.Add(new DoorInfo(x,y,(Door)e));
                    }
                }
            }

            return doors;
        } 

        private void RefreshLinkList()
        {
            List<DoorInfo> linkedDoors = new List<DoorInfo>();
            List<DoorInfo> unlinkedDoors = new List<DoorInfo>();

            foreach (DoorInfo door in editingDoors)
            {
                if (editingTrigger.Doors.Contains(door.Position))
                {
                    linkedDoors.Add(door);
                }
                else
                {
                    unlinkedDoors.Add(door);
                }
            }

            linkedDoorListBox.Items.Clear();
            foreach (DoorInfo door in linkedDoors)
            {
                linkedDoorListBox.Items.Add(door);
            }

            unlinkedDoorListBox.Items.Clear();
            foreach (DoorInfo door in unlinkedDoors)
            {
                unlinkedDoorListBox.Items.Add(door);
            }
        }

        private DoorInfo SelectedLinkDoor()
        {
            return unlinkedDoorListBox.SelectedItem as DoorInfo;
        }

        private DoorInfo SelectedUnlinkDoor()
        {
            return linkedDoorListBox.SelectedItem as DoorInfo;
        }

        private void buttonLink_Click(object sender, EventArgs e)
        {
            if(SelectedLinkDoor() != null)
            {
                editingTrigger.Link(SelectedLinkDoor().Position);
                RefreshLinkList();
            }
        }

        private void buttonUnlink_Click(object sender, EventArgs e)
        {
            if (SelectedUnlinkDoor() != null)
            {
                editingTrigger.UnLink(SelectedUnlinkDoor().Position);
                RefreshLinkList();
            }
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            this.DialogResult = System.Windows.Forms.DialogResult.OK;
            this.Close();
        }

        private void unlinkedDoorListBox_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (SelectedLinkDoor() != null)
            {
                editingTrigger.Link(SelectedLinkDoor().Position);
                RefreshLinkList();
            }
        }

        private void linkedDoorListBox_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (SelectedUnlinkDoor() != null)
            {
                editingTrigger.UnLink(SelectedUnlinkDoor().Position);
                RefreshLinkList();
            }
        }

        #endregion
        #region Edit Transport
        public ElementEditorDialog(Transport transport, IMazeMap map)
            : this()
        {
            tabControl.TabPages.Clear();
            tabControl.TabPages.Add(transportTabPage);

            editingTransport = transport;
            editingMap = map;

            InitializePassableNodes(map);
            UpdateTransportToState();
        }
        private void UpdateTransportToState()
        {
            try
            {
                transportLabel.Text =
                string.Format("{0} {1}",
                editingMap.GetElement(editingTransport.TransportTo).GetType().Name,
                editingTransport.TransportTo.ToString());
            }
            catch
            {
                try
                {
                    transportLabel.Text = editingTransport.TransportTo.ToString();
                }
                catch
                {
                    transportLabel.Text = "Unknown";
                }
            }
        }
        private void InitializePassableNodes(IMazeMap map)
        {
            int width = map.Width;
            int height = map.Height;

            //draw element
            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    MazeElement e = map.GetElement(x, y);
                    if (e != null && (e is Path || e is Entrance || e is Exit))
                    {
                        PositionType type = new PositionType(x, y, e);
                        passableListBox.Items.Add(type);
                    }
                }
            }
        }

        private void passableListBox_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            PositionType pt = passableListBox.SelectedItem as PositionType;

            if (editingTransport != null && pt != null)
            {
                editingTransport.TransportTo = pt.Position;
                UpdateTransportToState();
            }
        }  
        #endregion
    }
}
