﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.Engine;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Generators
{
    [Name("Lv1. Loop Maze")]
    [Author("GrapeCity")]
    public class LoopMazeGen : IMazeMapGenerator
    {
        public void SetupRenderSupport(Control screen)
        {
        }

        public void SetupSkinSupport(IMazeSkin skin)
        {
        }

        public bool Generate(int seed, IMazeMapEditor mapEditor)
        {
            GraphMazeGen.GenerateGraphMaze(seed, mapEditor, mapEditor.Width);

            MapUtil map = new MapUtil(mapEditor);

            int width = mapEditor.Width;
            int height = mapEditor.Height;

            for (int y = 0; y < height; y++)
            for (int x = 0; x < width; x++)
            {
                Position node = new Position(x, y);
                if(map.IsEntrance(node) || map.IsExit(node)) continue;

                if(CanConnect(map, node))
                {
                    map.SetElement(node, new Path(mapEditor));
                }
            }

            return true;
        }

        private static bool CanConnect(MapUtil map, IPosition current)
        {
            int closePathCount = 0;
            int cornerPathCount = 0;

            if (map.IsBoundary(current)) return false;

            IPosition[] list = map.CloseTo(current);
            foreach (IPosition check in list)
            {
                if (!map.IsBlock(check)) closePathCount++;
                if (map.IsExit(check)) closePathCount--;
            }

            IPosition[] cornerList = map.CornerTo(current);
            foreach (IPosition check in cornerList)
            {
                if (!map.IsBlock(check)) cornerPathCount++;
            }

            return closePathCount == 2 && cornerPathCount <= 3;
        }
    }

    [Name("Lv3. Grid Maze")]
    [Author("GrapeCity")]
    public class GridMazeGen : IMazeMapGenerator
    {
        public void SetupRenderSupport(Control screen)
        {
        }

        public void SetupSkinSupport(IMazeSkin skin)
        {
        }

        public bool Generate(int seed, IMazeMapEditor mapEditor)
        {
            GraphMazeGen.GenerateGraphMaze(seed, mapEditor, 0);
            return true;
        }
    }

    [Name("Lv2. Straight Maze")]
    [Author("GrapeCity")]
    public class GraphMazeGen : IMazeMapGenerator
    {
        public void SetupRenderSupport(Control screen)
        {
        }

        public void SetupSkinSupport(IMazeSkin skin)
        {
        }

        public bool Generate(int seed, IMazeMapEditor mapEditor)
        {
            GenerateGraphMaze(seed, mapEditor, mapEditor.Width / 2);
            return true;
        }

        internal static void GenerateGraphMaze(int seed, IMazeMapEditor mapEditor, int pathLength)
        {
            (new WallGen()).Generate(0, mapEditor);

            RandomUtil rnd = new RandomUtil(seed);
            MapUtil map = new MapUtil(mapEditor);

            //building graph nodes.
            Position[,] graphNodes = GridGen.GetGraphNodes(map.Width, map.Height);

            //hole graph nodes.
            int width = graphNodes.GetLength(0);
            int height = graphNodes.GetLength(1);
            bool[,] graphAccess = new bool[width, height];

            for (int y = 0; y < height; y++)
                for (int x = 0; x < width; x++)
                {
                    graphAccess[x, y] = false;
                    Position node = graphNodes[x, y];
                    if (map.IsExit(node) || map.IsEntrance(node)) continue;
                    map.SetElement(node, new Path(mapEditor));
                }

            //walk graph nodes.
            List<Position> accessList = new List<Position>();

            Position current = new Position(rnd.Next(0, width), rnd.Next(0, height));
            graphAccess[current.X, current.Y] = true;

            while (true)
            {
                int currentPathLength = 0;
                Direction lastDirection = Direction.Up;

                while (true)
                {
                    Direction[] valids = GetValidDirections(current, graphAccess, width, height);

                    if (valids != null && valids.Length > 0)
                    {
                        if (valids.Length > 1) accessList.Add(current);

                        Direction direction;

                        if (currentPathLength == 0)
                        {
                            direction = rnd.NextDirection(valids);
                            lastDirection = direction;
                        }
                        else if (currentPathLength > 0 && currentPathLength < pathLength)
                        {
                            bool havePath = false;

                            foreach (Direction valid in valids)
                            {
                                if (valid == lastDirection)
                                {
                                    havePath = true;
                                    break;
                                }
                            }

                            if (havePath)
                            {
                                direction = lastDirection;
                            }
                            else
                            {
                                direction = rnd.NextDirection(valids);
                                lastDirection = direction;
                            }
                        }
                        else
                        {
                            currentPathLength = 0;
                            direction = rnd.NextDirection(valids);
                            lastDirection = direction;
                        }

                        currentPathLength++;

                        if (direction != Direction.None)
                        {
                            Position walkTo = Position.GetPositionByDirection(current, direction);
                            graphAccess[walkTo.X, walkTo.Y] = true;

                            //break wall between current and walkto
                            Position wall = Position.GetPositionByDirection(graphNodes[current.X, current.Y], direction);
                            if (map.IsValidPosition(wall)) map.SetElement(wall, new Path(mapEditor));

                            current = walkTo;
                        }
                    }
                    else
                    {
                        break;
                    }
                }

                accessList.RemoveAll(delegate(Position p) { return Position.Equality(p, current); });

                List<Position> tempList = new List<Position>();
                foreach (Position position in accessList)
                {
                    if (!IsDeadPosition(position, graphAccess, width, height))
                    {
                        tempList.Add(position);
                    }
                }

                accessList.Clear();
                accessList = tempList;

                if (accessList.Count != 0)
                {
                    //Position[] list = new Position[accessList.Count];
                    //accessList.Keys.CopyTo(list, 0);
                    current = rnd.NextPosition(accessList);
                }
                else
                {
                    break;
                }
            }
        }

        private static readonly int UpDir = (int) Direction.Up;
        private static readonly int LeftDir = (int)Direction.Left;
        private static readonly int RightDir = (int)Direction.Right;
        private static readonly int DownDir = (int)Direction.Down;

        private static bool IsDeadPosition(IPosition current, bool[,] graphAccess, int width, int height)
        {
            int ox = current.X;
            int oy = current.Y;


            for (int dir = 0; dir < 4; dir++)
            {
                int x = ox;
                int y = oy;

                if (dir == UpDir) y--;
                else if (dir == LeftDir) x--;
                else if (dir == RightDir) x++;
                else y++;

                if (x >= 0 && x < width && y >= 0 && y < height && !graphAccess[x, y])
                {
                    return false;
                }
            }
            return true;
        }

        private static Direction[] GetValidDirections(IPosition current, bool[,] graphAccess, int width, int height)
        {
            List<Direction> list = new List<Direction>();

            for (int dir = 0; dir < 4; dir++)
            {
                Position p = Position.GetPositionByDirection(current, (Direction)dir);
                if (IsValidPosition(p, graphAccess, width, height))
                    list.Add((Direction)dir);
            }

            return list.ToArray();
        }

        private static bool IsValidPosition(IPosition test, bool[,] graphAccess, int width, int height)
        {
            if(test.X >=0 && test.X < width)
            if(test.Y >=0 && test.Y < height)
            if (!graphAccess[test.X, test.Y])
            {
                return true;
            }
            return false;
        }
    }
}
