﻿using System;
using System.Collections.Generic;
using System.Text;
using GrapeCity.Competition.Maze.Engine;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Generators
{
    internal class MapUtil : IMazeMapEditor
    {
        private readonly IMazeMapEditor map;

        public MapUtil(IMazeMapEditor map)
        {
            this.map = map;
        }

        public void SetElement(IPosition position, MazeElement element)
        {
            map.SetElement(position, element);
        }

        public void SetElement(int x, int y, MazeElement element)
        {
            map.SetElement(x, y, element);
        }

        public bool IsValidPosition(IPosition position)
        {
            return map.IsValidPosition(position);
        }

        public bool IsValidPosition(int x, int y)
        {
            return map.IsValidPosition(x, y);
        }

        public MazeElement GetElement(IPosition position)
        {
            return map.GetElement(position);
        }

        public MazeElement GetElement(int x, int y)
        {
            return map.GetElement(x, y);
        }

        public string Name
        {
            get { return map.Name; }
        }

        public int Width
        {
            get { return map.Width; }
        }

        public int Height
        {
            get { return map.Height; }
        }

        public IPosition Entrance
        {
            get { return map.Entrance; }
        }

        public IPosition Exit
        {
            get { return map.Exit; }
        }

        public bool IsEntrance(IPosition position)
        {
            return Position.Equality(map.Entrance, position);
        }

        public bool IsExit(IPosition position)
        {
            return Position.Equality(map.Exit, position);
        }

        public bool IsEditablePosition(IPosition position)
        {
            if (!map.IsValidPosition(position)) return false;
            if (Position.Equality(position, map.Entrance)) return false;
            if (Position.Equality(position, map.Exit)) return false;
            return true;
        }

        public bool IsBlock(int x, int y)
        {
            if (x < 0 || x >= map.Width) return true;
            if (y < 0 || y >= map.Height) return true;

            MazeElement e = map.GetElement(x, y);
            if (e == null) return true;
            if (e is Wall) return true;
            if (e is Door && ((Door)e).IsLocked) return true;

            return false;
        }

        public bool IsBlock(IPosition position)
        {
            return IsBlock(position.X, position.Y);
        }

        public void FillWalls()
        {
            int width = map.Width;
            int height = map.Height;

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    if (x == map.Entrance.X && y == map.Entrance.Y) continue;
                    if (x == map.Exit.X && y == map.Exit.Y) continue;

                    map.SetElement(x, y, new Wall(map));
                }
            }
        }

        public IPosition[] CloseTo(IPosition position)
        {
            List<IPosition> list = new List<IPosition>();

            for(int dir = 0; dir<4; dir++)
            {
                Position p = Position.GetPositionByDirection(position, (Direction) dir);
                if(map.IsValidPosition(p))
                    list.Add(p);
            }

            return list.ToArray();
        }

        public IPosition[] SurroundTo(IPosition position)
        {
            List<IPosition> list = new List<IPosition>();

            for (int dir = 0; dir < 4; dir++)
            {
                Position p = Position.GetPositionByDirection(position, (Direction)dir);
                if (map.IsValidPosition(p)) list.Add(p);
            }

            Position p_lu = new Position(position.X - 1, position.Y - 1);
            if (map.IsValidPosition(p_lu)) list.Add(p_lu);

            Position p_ld = new Position(position.X - 1, position.Y + 1);
            if (map.IsValidPosition(p_ld)) list.Add(p_ld);

            Position p_ru = new Position(position.X + 1, position.Y - 1);
            if (map.IsValidPosition(p_ru)) list.Add(p_ru);

            Position p_rd = new Position(position.X + 1, position.Y + 1);
            if (map.IsValidPosition(p_rd)) list.Add(p_rd);

            return list.ToArray();
        }

        public IPosition[] CornerTo(IPosition position)
        {
            List<IPosition> list = new List<IPosition>();

            Position p_lu = new Position(position.X - 1, position.Y - 1);
            if (map.IsValidPosition(p_lu)) list.Add(p_lu);

            Position p_ld = new Position(position.X - 1, position.Y + 1);
            if (map.IsValidPosition(p_ld)) list.Add(p_ld);

            Position p_ru = new Position(position.X + 1, position.Y - 1);
            if (map.IsValidPosition(p_ru)) list.Add(p_ru);

            Position p_rd = new Position(position.X + 1, position.Y + 1);
            if (map.IsValidPosition(p_rd)) list.Add(p_rd);

            return list.ToArray();
        }

        public Direction[] GetValidDirections(IPosition position)
        {
            List<Direction> list = new List<Direction>();

            for (int dir = 0; dir < 4; dir++)
            {
                Position p = Position.GetPositionByDirection(position, (Direction)dir);
                if (map.IsValidPosition(p))
                    list.Add((Direction)dir);
            }

            return list.ToArray();
        }

        public Direction[] GetValidDirections(IPosition position, IPosition ignore)
        {
            List<Direction> list = new List<Direction>();

            for (int dir = 0; dir < 4; dir++)
            {
                Position p = Position.GetPositionByDirection(position, (Direction)dir);
                if (map.IsValidPosition(p) && !Position.Equality(p, ignore))
                    list.Add((Direction)dir);
            }

            return list.ToArray();
        }

        public bool IsBoundary(IPosition position)
        {
            if (position.X == 0 || position.Y == 0) return true;
            else if (position.X == map.Width - 1 || position.Y == map.Height - 1) return true;
            return false;
        }
    }
}
