﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.Engine;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Maps;

namespace GrapeCity.Competition.Maze.Generators
{
    public partial class MazeEditorDialog : Form
    {
        private readonly IMazeMapEditor m_map;
        private readonly IMazeSkin m_skin;
        private int m_blockSize;
        private bool m_dirty;
        private bool m_painting;
        private Image m_screenCache;
        private Position m_mousePosition = Position.Zero;
        private MazeElement m_mouseElement = null;
        private const string m_stateFormat = "Current: {0} \tID: {1}";
        private bool m_mouseLeftDown = false;

        private IMazeMapEditor Map
        {
            get { return m_map; }
        }

        private IMazeSkin Skin
        {
            get { return m_skin; }
        }

        #region Render Screen
        public MazeEditorDialog(int seed, IMazeMapEditor map, IMazeSkin skin)
        {
            m_map = map;
            m_skin = skin;
            m_dirty = true;
            m_painting = false;

            InitializeComponent();

            seedToolStripTextBox.Text = seed.ToString();
            pathToolStripButton.Tag = new Path(map);
            wallToolStripButton.Tag = new Wall(map);
            triggerEditorToolStripButton.Tag = new Trigger(map);
            doorToolStripButton.Tag = new Door(map);
            transportToolStripButton.Tag = new Transport(map);
        }

        private void InitializeGeneratorList()
        {
            Assembly container = typeof(IMazeMapGenerator).Assembly;

            List<NamedType> generators = new List<NamedType>();
            foreach (Type type in container.GetTypes())
            {
                if (type.GetInterface("IMazeMapGenerator") == typeof(IMazeMapGenerator))
                {
                    generators.Add(new NamedType(type));
                }
            }

            generatorToolStripComboBox.Items.AddRange(generators.ToArray());
            if (generatorToolStripComboBox.Items.Count > 0)
                generatorToolStripComboBox.SelectedIndex = 0;
        }

        private void ForceUpdate()
        {
            this.m_dirty = true;
            this.editorScreen.Refresh();
        }

        private void MazeEditorDialog_Load(object sender, EventArgs e)
        {
            InitializeGeneratorList();
            MazeEditorDialog_Resize(sender, e);
        }

        private void randomToolStripButton_Click(object sender, EventArgs e)
        {
            seedToolStripTextBox.Text = (new Random().Next()).ToString();
            seedToolStripButton_Click(sender, e);
        }

        private void seedToolStripButton_Click(object sender, EventArgs e)
        {
            NamedType gen = generatorToolStripComboBox.SelectedItem as NamedType;
            if (gen != null)
            {
                int seed;
                if (int.TryParse(seedToolStripTextBox.Text, out seed))
                {
                    try
                    {
                        IMazeMapGenerator generator = Activator.CreateInstance(gen.Type) as IMazeMapGenerator;
                        if (generator != null)
                        {
                            generator.SetupSkinSupport(Skin);
                            generator.Generate(seed, Map);
                            ForceUpdate();
                        }
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(this, "Can not load maze generator. because " + ex.Message);
                    }
                }
            }
        }

        private void returnToolStripButton_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void editorScreen_Paint(object sender, PaintEventArgs e)
        {
            #region Paint
            try
            {
                m_painting = true;


                if (m_dirty)
                {
                    m_screenCache = new Bitmap(editorScreen.Width, editorScreen.Height);
                    Graphics cache = Graphics.FromImage(m_screenCache);

                    int width = Map.Width;
                    int height = Map.Height;

                    //draw element
                    for (int y = 0; y < height; y++)
                    {
                        for (int x = 0; x < width; x++)
                        {
                            MazeElement element = Map.GetElement(x, y);
                            if (Skin != null) Skin.PaintElement(cache, x, y, m_blockSize, element);
                        }
                    }

                    //draw relation
                    for (int y = 0; y < height; y++)
                    {
                        for (int x = 0; x < width; x++)
                        {
                            MazeElement element = Map.GetElement(x, y);

                            if (element is Transport)
                            {
                                Transport transport = (Transport)element;
                                LineTo(cache, Color.DarkBlue, x, y, transport.TransportTo);
                            }

                            if (element is Trigger)
                            {
                                Trigger trigger = (Trigger)element;

                                foreach (Position door in trigger.Doors)
                                {
                                    LineTo(cache, Color.DarkGreen, x, y, door);
                                }
                            }
                        }
                    }

                    m_dirty = false;
                }

                if (m_screenCache != null)
                {
                    e.Graphics.DrawImage(m_screenCache, 0, 0);
                }
            }
            finally 
            {
                m_painting = false;
            }
            #endregion
        }

        private void LineTo(Graphics g, Color color, int x, int y, IPosition to)
        {
            if(m_blockSize > 1)
            {
                int offset = (int)(0.5*m_blockSize);

                Pen pen = new Pen(color, 2);
                pen.CustomEndCap =  new System.Drawing.Drawing2D.AdjustableArrowCap((float)(pen.Width * 3), (float)(pen.Width * 4), true);
                
                g.DrawLine(pen,
                        x * m_blockSize + offset,
                        y * m_blockSize + offset,
                        to.X * m_blockSize + offset,
                        to.Y * m_blockSize + offset);
            }
            else
            {
                g.DrawLine(new Pen(color, 2),
                        x * m_blockSize,
                        y * m_blockSize,
                        to.X * m_blockSize,
                        to.Y * m_blockSize);
            }
        }

        private void MazeEditorDialog_Resize(object sender, EventArgs e)
        {
            try
            {
                //calc block size
                int blockWidth;
                int blockHeight;

                if (blockSizeTrackBar.Value > 64)
                {
                    blockWidth = panel1.Width / Map.Width;
                    blockHeight = panel1.Height / Map.Height;
                }
                else
                {
                    blockHeight = blockWidth = blockSizeTrackBar.Value;
                }

                int newSize = Math.Max(Math.Min(blockWidth, blockHeight), 1);
                //new block size

                if (m_blockSize != newSize)
                {
                    m_dirty = true;
                    m_blockSize = newSize;
                }

                //update screen size
                editorScreen.Width = Map.Width * m_blockSize;
                editorScreen.Height = Map.Height * m_blockSize;
                this.editorScreen.Refresh();
            }
            catch
            {
            }
        }

        private void blockSizeTrackBar_ValueChanged(object sender, EventArgs e)
        {
            if (!m_painting)
            {
                MazeEditorDialog_Resize(sender, e);
            }
        } 

        #endregion

        #region Update Button CheckState
        private void UncheckAllToolStripButton()
        {
            pathToolStripButton.Checked = false;
            wallToolStripButton.Checked = false;
            doorToolStripButton.Checked = false;
            triggerToolStripButton.Checked = false;
            transportToolStripButton.Checked = false;
        }

        private void elementToolStripButton_Click(object sender, EventArgs e)
        {
            UncheckAllToolStripButton();
            ToolStripButton button = sender as ToolStripButton;
            if (button != null)
            {
                button.Checked = true;
                m_mouseElement = button.Tag as MazeElement;
            }
        } 

        private MazeElement GetMouseElement()
        {
            if (wallToolStripButton.Checked) return new Wall(Map);
            else if (triggerToolStripButton.Checked) return new Trigger(Map);
            else if (doorToolStripButton.Checked) return new Door(Map);
            else if (transportToolStripButton.Checked) return new Transport(Map);
            else return new Path(Map);
        }
        #endregion

        private void UpdateStateBarInfo()
        {
            this.toolStripStatusLabel.Text = string.Format(m_stateFormat, m_mousePosition, "");
        }

        #region Element edit
        private void editorScreen_MouseMove(object sender, MouseEventArgs e)
        {
            int x = e.X / m_blockSize;
            int y = e.Y / m_blockSize;

            m_mousePosition = new Position(x, y);
            SetElement();

            UpdateStateBarInfo();
        }

        private void editorScreen_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left) m_mouseLeftDown = true;
            SetElement();
        }

        private void editorScreen_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left) m_mouseLeftDown = false;
        }

        private void SetElement()
        {
            if (Position.Equality(m_mousePosition, Map.Entrance)) return;
            if (Position.Equality(m_mousePosition, Map.Exit)) return;

            m_mouseElement = GetMouseElement();

            if (m_mouseLeftDown)
            {
                if(Map.IsValidPosition(m_mousePosition))
                {
                    Map.SetElement(m_mousePosition, m_mouseElement.Clone(Map));
                    ForceUpdate();
                }
            }
        }

        private void editorScreen_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == System.Windows.Forms.MouseButtons.Right)
            {
                MazeElement me = Map.GetElement(m_mousePosition);

                if (me is Door)
                {
                    ElementEditorDialog dialog = new ElementEditorDialog(me as Door);
                    dialog.ShowDialog(this);
                    ForceUpdate();
                }
                else if (me is Trigger)
                {
                    ElementEditorDialog dialog = new ElementEditorDialog(me as Trigger, Map);
                    dialog.ShowDialog(this);
                    ForceUpdate();
                }
                else if (me is Transport)
                {
                    ElementEditorDialog dialog = new ElementEditorDialog(me as Transport, Map);
                    dialog.ShowDialog(this);
                    ForceUpdate();
                }
            }
        } 
        #endregion

        private void MazeEditorDialog_KeyDown(object sender, KeyEventArgs e)
        {
            if(e.KeyCode == Keys.Escape)
            {
                UncheckAllToolStripButton();
                pathToolStripButton.Checked = true;
            }
            else if (e.KeyCode == Keys.F1)
            {
                UncheckAllToolStripButton();
                wallToolStripButton.Checked = true;
            }
            else if (e.KeyCode == Keys.F2)
            {
                UncheckAllToolStripButton();
                doorToolStripButton.Checked = true;
            }
            else if (e.KeyCode == Keys.F3)
            {
                UncheckAllToolStripButton();
                triggerToolStripButton.Checked = true;
            }
            else if (e.KeyCode == Keys.F4)
            {
                UncheckAllToolStripButton();
                transportToolStripButton.Checked = true;
            }
        }
    }
}
