﻿using System;

namespace GrapeCity.Competition.Maze
{
    #region The IExplorer interface

    /// <summary>
    ///   Represents the explorer in maze.
    /// </summary>
    /// <example>
    /// <code>  
    ///  public void Think(IMaze maze, IExplorer explorer)
    ///  {
    ///     Trace.WriteLine(explorer.Name + " is enter the maze.");
    ///     explorer.Say("Hello, everyone! i am " + explorer.Name, null);
    /// 
    ///     while(explorer.Move(Direction.Up))
    ///     {}
    /// 
    ///     while(explorer.Move(Direction.Left))
    ///     {}
    ///  }
    /// </code>
    /// </example>
    public interface IExplorer : IPosition
    {
        /// <summary>
        ///   Gets the explorer name.
        /// </summary>
        /// <value>
        ///   A <b>string</b> value indicates the explorer name.
        /// </value>
        string Name { get; }

        /// <summary>
        ///   Gets the total steps
        /// </summary>
        /// <value>
        ///   An <b>int</b> value indicates the total steps.
        /// </value>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     while(explorer.Steps &lt;= 100)
        ///     {         
        ///         explorer.Move((Direction)(new Random()).Next(0,4));
        ///     } 
        ///  }
        /// </code>
        /// </example>
        int Steps { get; }

        /// <summary>
        ///   Gets the information of current position.
        /// </summary>
        /// <value>
        ///   A <see cref="PositionInfo"/> value indicates the information of current position.
        /// </value>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     while(explorer.Current != PositionInfo.Exit)
        ///     {
        ///         while(explorer.Move(Direction.Up){}
        ///         while(explorer.Move(Direction.Left){}
        ///     }
        ///     Trace.WriteLine(explorer.Current);
        ///  }
        /// </code>
        /// </example>
        PositionInfo Current { get; }

        /// <summary>
        ///   Gets the current position.
        /// </summary>
        /// <value>
        ///   A <see cref="IPosition"/> value indicates the position.
        /// </value>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     while(explorer.Location.X != maze.Exit.X || explorer.Location.Y != maze.Exit.Y)
        ///     {
        ///         //...
        ///     }
        ///  }
        /// </code>
        /// </example>
        IPosition Location { get; }

        /// <summary>
        ///   Gets the information of surrounding position. 
        /// </summary>
        /// <param name="direction">
        ///   A <see cref="Direction"/> value indicates the surrounding position. 
        /// </param>
        /// <returns>
        ///   A <see cref="PositionInfo"/> value indicates the information of surrounding position.
        /// </returns>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     if(explorer.Look(Direction.Left) == PositionInfo.Path)
        ///     {
        ///         explorer.Move(Direction.Left);
        ///     }
        ///  }
        /// </code>
        /// </example>
        PositionInfo Look(Direction direction);

        /// <summary>
        ///   Say something to other explorer in the maze.
        /// </summary>
        /// <param name="message">
        ///   A <b>string</b> value indicates the message.
        /// </param>
        /// <param name="information">
        ///   An <see cref="System.Object"/> indicates the extra information.
        /// </param>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     explorer.Say("Hello, Maze!", null);
        ///  }
        /// </code>
        /// </example>
        void Say(string message, object information);

        /// <summary>
        ///   Whisper something to other explorer in the maze.
        /// </summary>
        /// <param name="name">
        ///   A <b>string</b> value indicates the listener's name.
        /// </param>
        /// <param name="message">
        ///   A <b>string</b> value indicates the message.
        /// </param>
        /// <param name="information">
        ///   An <see cref="System.Object"/> indicates the extra information.
        /// </param>
        /// <returns>
        ///  <b>true</b>, if the operation is successful; otherwise, <b>false</b>.
        /// </returns>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     explorer.Wisper("Peter", "Where are you?", null);
        ///  }
        /// </code>
        /// </example>
        bool Whisper(string name, string message, object information);

        /// <summary>
        ///   Stay in current position. 
        /// </summary>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     if(explorer.Current == PositionInfo.Trigger)
        ///     {
        ///         explorer.Stay();
        ///     }     
        ///  }
        /// </code>
        /// </example>
        void Stay();

        /// <summary>
        ///   Move to next position.
        /// </summary>
        /// <param name="direction">
        ///   A <see cref="Direction"/> value indicates the moving direction.
        /// </param>
        /// <returns>
        ///   <b>true</b>, if the moving is successful; otherwise, <b>false</b>.
        /// </returns>
        /// <example>
        /// <code>  
        ///  public void Think(IMaze maze, IExplorer explorer)
        ///  {
        ///     if(explorer.Look(Direction.Left) == PositionInfo.Path)
        ///     {
        ///         explorer.Move(Direction.Left);
        ///     }
        ///  }
        /// </code>
        /// </example>
        bool Move(Direction direction);
    }
    #endregion
}
