﻿using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.Engine;
using GrapeCity.Competition.Maze.Engine.Elements;
using GrapeCity.Competition.Maze.Engine.Explorers;
using GrapeCity.Competition.Maze.Engine.Maps;
using GrapeCity.Competition.Maze.Generators;

namespace GrapeCity.Competition.Maze.EmbeddedEngine
{
    internal class GraphicsEngine : MazeGame
    {
        private Image screenCache;
        private int screenBlockSize;

        private Control screen;
        private IMazeSkin mazeSkin;
        private IEngineRecorder recorder;

        private Queue<Position> explorerFootprint;
        private List<Position> doorList;

        private int footprintLenght;

        internal GraphicsEngine(string name, int width, int height)
            : this(new MazeMap(name, width, height))
        {
        }

        internal GraphicsEngine(MazeMap maze)
            : base(maze)
        {
            explorerFootprint = new Queue<Position>();
            doorList = new List<Position>();
            footprintLenght = (Map.Width + Map.Height);
        }

        internal MazeMap MazeMap
        {
            get { return base.StaticMap; }
        }

        internal IEngineRecorder Recorder
        {
            get { return this.recorder; }
            set { this.recorder = value; }
        }

        internal Control Screen
        {
            get { return screen; }
            set
            {
                if(screen != value && value != null)
                {
                    screen = value;
                    screen.Paint += PaintMaze;
                }
            }
        }

        internal IMazeSkin Skin
        {
            get { return mazeSkin; }
            set { mazeSkin = value; }
        }

        protected override void OnGameStart()
        {
            explorerFootprint.Clear();

            int width = Map.Width;
            int height = Map.Height;

            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    MazeElement element = Map.GetElement(x, y);
                    if(element is Door)
                    {
                        doorList.Add(new Position(x,y));
                    }
                }
            }
        }

        protected override void OnExplorerSay(MazeExplorer e, string message, object information)
        {
            if (recorder != null) recorder.WriteRecord(e.Name, "Says: " + message);
            base.OnExplorerSay(e, message, information);
        }

        protected override bool OnExplorerWisper(MazeExplorer e, string name, string message, object information)
        {
            if (recorder != null) recorder.WriteRecord(e.Name, string.Format("Wisper to {0}: {1}", name, message));
            return base.OnExplorerWisper(e, name, message, information);
        }

        protected override void OnRender()
        {
            if (screen != null) screen.Refresh();
            Application.DoEvents();
        }

        protected override void OnGameStop()
        {
            explorerFootprint.Clear();
        }

        protected override bool OnExplorerMove(MazeExplorer e, Direction direction)
        {
            Position p = new Position(e);
            bool moved = base.OnExplorerMove(e, direction);

            if(moved && direction != Direction.None)
            {
                lock (explorerFootprint)
                {
                    explorerFootprint.Enqueue(p);
                    if (explorerFootprint.Count > footprintLenght) explorerFootprint.Dequeue();
                }
            }

            return moved;
        }

        protected virtual void PaintMaze(object sender, PaintEventArgs e)
        {
            if(mazeSkin == null) return;
            Graphics g = e.Graphics;

            UpdateScreenCache();
            PaintFromCache(g);

            //draw doors

            foreach (Position position in doorList)
            {
                MazeElement door = Map.GetElement(position);
                mazeSkin.PaintElement(g, position.X, position.Y, door);
            }

            //draw footprint
            Position[] footprints;

            lock (explorerFootprint)
            {
                footprints = explorerFootprint.ToArray();
            }

            foreach (Position position in footprints)
            {
                if(mazeSkin != null)
                {
                    mazeSkin.PaintFootPrint(g, position.X, position.Y);
                }
            }

            //draw explorers
            foreach (KeyValuePair<string, MazeExplorer> pair in this.Explorers)
            {
                if (mazeSkin != null)
                {
                    mazeSkin.PaintExplorer(g, pair.Value);
                }
            }
        }

        internal void ForceUpdate()
        {
            screenBlockSize = 0;
            InitialReachs();
        }

        private void UpdateScreenCache()
        {
            if (screenCache != null && mazeSkin != null && screenBlockSize == mazeSkin.BlockSize) return;

            int width = Map.Width;
            int height = Map.Height;

            screenBlockSize = mazeSkin.BlockSize;
            screen.Width = width * mazeSkin.BlockSize;
            screen.Height = height * mazeSkin.BlockSize;

            screenCache = new Bitmap(screen.Width, screen.Height);
            Graphics cache = Graphics.FromImage(screenCache);

            //draw element
            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    if (mazeSkin != null)
                    {
                        mazeSkin.PaintElement(cache, x, y, Map.GetElement(x, y));
                    }
                }
            }
        }

        private void PaintFromCache(Graphics g)
        {
            g.DrawImage(screenCache,0,0);
        }

        internal void GenerateMap(IMazeMapGenerator gen, int seed)
        {
            if(gen != null)
            {
                if (Running)
                {
                    gen.Generate(seed, this.Map);
                }
                else
                {
                    gen.Generate(seed, this.StaticMap);
                    ResetMap();
                }

                ForceUpdate();
            }
        }

        internal void Shutdown()
        {
            screen.Paint -= PaintMaze;
        }
    }
}
