﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.EmbeddedEngine;
using GrapeCity.Competition.Maze.Engine.Maps;
using GrapeCity.Competition.Maze.Generators;

namespace GrapeCity.Competition.Maze
{
    internal class MazeGameConsole
    {
        #region Fields
        private readonly MazeSettings settings;
        private readonly MazeSkin skin;

        private readonly Control screen;
        private readonly List<string> assemblies;
        private readonly List<string> recentMaps;
        private readonly List<MazePlayer> players;

        private GraphicsEngine game;
        private IEngineRecorder recorder;

        private string mazeFilename; 

        #endregion
        #region Constructor
        /// <summary>
        /// Initialize a maze game console with a screen control
        /// </summary>
        /// <param name="ScreenWinControl"></param>
        internal MazeGameConsole(Control ScreenWinControl)
        {
            this.screen = ScreenWinControl;

            this.skin = new MazeSkin();
            this.assemblies = new List<string>();
            this.recentMaps = new List<string>();
            this.players = new List<MazePlayer>();
            this.settings = new MazeSettings();

            ApplySettings();
            Application.ApplicationExit += Application_ApplicationExit;
        } 
        #endregion
        #region Properties

        internal string MapFileName
        {
            get { return mazeFilename; }
        }

        /// <summary>
        /// The maze skin.
        /// </summary>
        internal MazeSkin Skin
        {
            get { return this.skin; }
        }

        /// <summary>
        /// Gets assembly files.
        /// </summary>
        internal List<string> Assemblies
        {
            get { return assemblies; }
        }

        /// <summary>
        /// Gets recent maps.
        /// </summary>
        internal List<string> RecentMaps
        {
            get { return recentMaps; }
        }

        /// <summary>
        /// Gets the players.
        /// </summary>
        internal List<MazePlayer> Players
        {
            get { return players; }
        }

        /// <summary>
        /// Gets the maze engine status.
        /// </summary>
        internal MazeStatus Status
        {
            get
            {
                if (game != null)
                    return new MazeStatus(game.Name, game.Width, game.Height, mazeFilename, game.Exit, game.ReachBlocks, game.TotalBlocks, game.MazeExplorers);
                else
                    return new MazeStatus(mazeFilename);
            }
        } 
        #endregion
        #region Create Maze
        internal bool CanCreateMaze
        {
            get { return this.game == null || !this.game.Running; }
        }

        internal void CreateMaze(int height, int width, string name)
        {
            if (game != null) game.Shutdown();

            game = new GraphicsEngine(name, width, height);
            game.Screen = this.screen;
            game.Skin = this.skin;
            game.Recorder = this.recorder;
            mazeFilename = string.Empty;
        } 

        /// <summary>
        /// Create a blank maze game.
        /// </summary>
        /// <param name="height"></param>
        /// <param name="width"></param>
        /// <param name="name"></param>
        /// <param name="entrance"></param>
        /// <param name="exit"></param>
        internal void CreateMaze(int height, int width, string name, IPosition entrance, IPosition exit)
        {
            if(game!= null) game.Shutdown();

            MazeMap map = new MazeMap(name, width, height, entrance, exit);

            game = new GraphicsEngine(map);
            game.Screen = this.screen;
            game.Skin = this.skin;
            game.Recorder = this.recorder;
            mazeFilename = string.Empty;
        } 
        #endregion
        #region Load Maze
        internal bool CanLoadMaze
        {
            get { return this.game == null || !this.game.Running; }
        }

        /// <summary>
        /// Create a maze game with an existing maze map.
        /// </summary>
        /// <param name="filename"></param>
        internal bool LoadMaze(string filename)
        {
            MazeMap loadMap;

            try
            {
                IFormatter formatter = new BinaryFormatter();
                Stream stream = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.Read);
                loadMap = (MazeMap)formatter.Deserialize(stream);
                stream.Close();
            }
            catch
            {
                return false;
            }


            if (game != null) game.Shutdown();
            game = new GraphicsEngine(loadMap);
            game.Screen = this.screen;
            game.Skin = this.skin;
            game.Recorder = this.recorder;
            mazeFilename = filename;
            return true;
        }
        #endregion
        #region Save Maze
        internal bool CanSaveMaze
        {
            get { return this.game != null; }
        }

        /// <summary>
        /// Save the maze map.
        /// </summary>
        internal bool SaveMaze()
        {
            if (string.IsNullOrEmpty(mazeFilename))
                return false;

            SaveMaze(mazeFilename);

            return true;
        }

        /// <summary>
        /// Save the maze map to a file.
        /// </summary>
        /// <param name="name"></param>
        internal void SaveMaze(string name)
        {
            mazeFilename = name;

            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(name, FileMode.Create, FileAccess.Write, FileShare.None);
            formatter.Serialize(stream, game.MazeMap);
            stream.Close();
        }

        #endregion
        #region Run maze

        internal bool CanRun
        {
            get
            {
                return game != null && !game.Running;
            }
        }

        internal bool NeedConfigExplorer
        {
            get { return players.Count == 0; }
        }

        internal void Run(int delay)
        {
            foreach (MazePlayer player in players){
                game.CreateExplorer(player.Name, player.AIType);
            }

            game.Interval = delay;
            game.Skin = skin;
            game.Screen = screen;
            game.Start();
        }

        #endregion
        #region Pause Maze
        internal bool CanPause
        {
            get { return game != null && game.Running; }
        }

        internal void Pause()
        {
            game.Paused = !game.Paused;
        } 
        #endregion
        #region Stop Maze
        internal bool CanStop
        {
            get { return game != null && game.Running; }
        }

        internal void Stop()
        {
            game.Stop();
            game.Reset();
            game.ForceUpdate();
        } 

        #endregion
        #region Generate Maze
        internal bool CanGenerateMaze
        {
            get { return this.game != null && !this.game.Running; }
        }

        internal void GenerateMaze(int seed, Type generator)
        {
            try
            {
                IMazeMapGenerator mapGenerator = Activator.CreateInstance(generator) as IMazeMapGenerator;

                if(mapGenerator != null)
                {
                    mapGenerator.SetupRenderSupport(screen);
                    mapGenerator.SetupSkinSupport(skin);

                    game.GenerateMap(mapGenerator, seed);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show("Can not load maze generator. because " + e.Message + "\r\nTrace:" + e.StackTrace);
            }
        } 
        #endregion
        #region Can Configuration explorer
        internal bool CanConfigExplorer
        {
            get
            {
                return game == null || !game.Running;
            }
        } 
        #endregion
        #region Resize Screen
        internal void AdjustScreenSize(int width, int height)
        {
            if (game == null) return;
            //adjust block size
            int blockWidth = width / game.Width;
            int blockHeight = height / game.Height;
            this.skin.BlockSize = Math.Max(Math.Min(blockWidth, blockHeight),1);
            this.screen.Refresh();
        }
        internal void Update()
        {
            this.game.ForceUpdate();
        }
        #endregion
        #region Recorder
        internal IEngineRecorder Recorder
        {
            get { return recorder; }
            set 
            { 
                recorder = value;
                if (recorder != null && game != null) game.Recorder = value;
            }
        } 
        #endregion
        #region Application Settings

        void ApplySettings()
        {
            skin.SkinColors = settings.SkinColors;
            if(settings.Assemblies != null) assemblies.AddRange(settings.Assemblies);
            if(settings.RecentMaps != null) recentMaps.AddRange(settings.RecentMaps);
        }

        public void SaveSettings()
        {
            settings.SkinColors = skin.SkinColors;
            settings.Assemblies = assemblies;
            settings.RecentMaps = recentMaps;
            settings.Save();
        }

        void Application_ApplicationExit(object sender, EventArgs e)
        {
            SaveSettings();
        }

        #endregion
    }
}