﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.Utils;

namespace GrapeCity.Competition.Maze.UI
{
    internal partial class ConfigurateExplorerDialog : Form
    {
        #region The MazeAI class
        private class MazeAI
        {
            private readonly Type innerType;
            private readonly string name;
            private readonly string author;
            private readonly string assemblyName;
            private readonly string version;

            internal MazeAI(Type type)
            {
                if (type == null) throw new ArgumentNullException("type");

                this.innerType = type;
                this.name = NameAttribute.GetName(type);
                this.author = AuthorAttribute.GetName(type);
                this.assemblyName = type.Assembly.GetName().Name;
                this.version = type.Assembly.GetName().Version.ToString();
            }

            internal Type Type
            {
                get { return innerType; }
            }

            internal string Version
            {
                get { return version; }
            }

            internal string Name
            {
                get { return name; }
            }

            internal string Author
            {
                get { return author; }
            }

            internal string AssemblyName
            {
                get { return assemblyName; }
            }
        } 
        #endregion

        private readonly MazeGameConsole console;

        internal ConfigurateExplorerDialog(MazeGameConsole console)
        {
            InitializeComponent();

            this.console = console;
            groupByComboBox.Text = groupByComboBox.Items[0].ToString();
        }

        private void ConfigurateExplorerDialog_Load(object sender, EventArgs e)
        {
            RefreshPlayerList();

            if(console.Assemblies == null || console.Assemblies.Count == 0)
            {
                AssemblyLibraryDialog dialog = new AssemblyLibraryDialog(this.console.Assemblies);
                dialog.ShowDialog(this);
            }

            RefreshMazeAIList();
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.OK;
            Close();
        }

        private void groupByComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            RefreshMazeAIList();
        }

        private void buttonAssemblyExplorer_Click(object sender, EventArgs e)
        {
            AssemblyLibraryDialog dialog = new AssemblyLibraryDialog(this.console.Assemblies);
            dialog.ShowDialog(this);
            RefreshMazeAIList();
        }

        #region RefreshMazeAIList()
        private void RefreshMazeAIList()
        {
            //grouping
            bool groupByAuthor = groupByComboBox.Text == "Author";
            //bool groupByAssembly = !groupByAuthor;

            List<MazeAI> list = new List<MazeAI>();
            #region Init MazeAI
            foreach (string filename in console.Assemblies)
            {
                List<MazeAI> part = LoadMazeAIFromAssembly(filename);
                if (part.Count > 0) list.AddRange(part);
            }

            //sort
            list.Sort(ListComparer<MazeAI>.FromProperty("Name"));
            #endregion

            Dictionary<string, ListViewGroup> groups = new Dictionary<string, ListViewGroup>();
            #region Init Groups
            if (groupByAuthor)
            {
                foreach (MazeAI mazeAI in list)
                {
                    if (!groups.ContainsKey(mazeAI.Author))
                        groups.Add(mazeAI.Author, new ListViewGroup(mazeAI.Author));
                }
            }
            else //groupByAssembly
            {
                foreach (MazeAI mazeAI in list)
                {
                    if (!groups.ContainsKey(mazeAI.AssemblyName))
                        groups.Add(mazeAI.AssemblyName, new ListViewGroup(mazeAI.AssemblyName));
                }
            }

            #endregion

            #region create list view group
            List<ListViewGroup> groupList = new List<ListViewGroup>(groups.Values);
            groupList.Sort(ListComparer<ListViewGroup>.FromProperty("Header"));

            assemblyListView.Groups.Clear();
            assemblyListView.Groups.AddRange(groupList.ToArray());
            #endregion

            //create list view item
            #region create list view item
            assemblyListView.Items.Clear();
            foreach (MazeAI ai in list)
            {
                ListViewItem item = new ListViewItem();
                item.Tag = ai.Type;
                item.Text = ai.Name;
                item.SubItems.Add(ai.Version);
                item.SubItems.Add(ai.Author);
                item.SubItems.Add(ai.AssemblyName);

                if (groupByAuthor)
                    item.Group = groups[ai.Author];
                else
                    item.Group = groups[ai.AssemblyName];

                assemblyListView.Items.Add(item);
            }
            #endregion

            assemblyListView.AutoResizeColumns(ColumnHeaderAutoResizeStyle.HeaderSize);
        } 
        #endregion

        #region RefreshPlayerList()
        private void RefreshPlayerList()
        {
            //create list view item
            #region create list view item
            explorerListView.Items.Clear();

            foreach (MazePlayer player in console.Players)
            {
                ListViewItem item = new ListViewItem();
                item.Tag = player;
                item.Text = player.Name;
                item.SubItems.Add(player.AIName);
                item.SubItems.Add(player.Version);
                item.SubItems.Add(player.AIAuthor);
                item.SubItems.Add(player.AIType.FullName);

                explorerListView.Items.Add(item);
            }
            #endregion

            explorerListView.AutoResizeColumns(ColumnHeaderAutoResizeStyle.HeaderSize);
        } 
        #endregion

        #region LoadMazeAIFromAssembly(string)
        private static List<MazeAI> LoadMazeAIFromAssembly(string filename)
        {
            List<MazeAI> mazeAiList = new List<MazeAI>();

            try
            {
                Assembly assembly = Assembly.LoadFrom(filename);

                foreach (Type type in assembly.GetTypes())
                {
                    if (type.GetInterface("IMazeAI") == typeof(IMazeAI))
                        mazeAiList.Add(new MazeAI(type));
                }
            }
            catch { }

            return mazeAiList;
        } 
        #endregion

        #region Create Player
        private Type GetSelectedAI()
        {
            if (assemblyListView.SelectedItems.Count > 0)
                return assemblyListView.SelectedItems[0].Tag as Type;
            return null;
        }

        private MazePlayer CreatePlayer()
        {
            Type type = GetSelectedAI();

            if (type == null) return null;

            string namePrefix = NameAttribute.GetName(type);
            int nameSuffix = 1;
            string explorerName = string.Format("{0}", namePrefix);
            do
            {
                bool contains = false;
                foreach (MazePlayer player in this.console.Players)
                {
                    if (player.Name == explorerName)
                    {
                        contains = true;
                        break;
                    }
                }
                if (!contains) break;

                explorerName = string.Format("{0}({1})", namePrefix, nameSuffix);
                nameSuffix++;

            } while (true);

            return new MazePlayer(explorerName, type);
        } 
        #endregion

        private void buttonCreateExplorer_Click(object sender, EventArgs e)
        {
            MazePlayer player = CreatePlayer();
            if (player != null)
            {
                if(this.console.Players.Count >= 3)
                {
                    MessageBox.Show("You can create 3 explorer at most.");
                    return;
                }

                this.console.Players.Add(player);
                RefreshPlayerList();
            }
        }

        private void buttonRemoveExplorer_Click(object sender, EventArgs e)
        {
            if(explorerListView.SelectedItems.Count > 0)
            {
                foreach (ListViewItem item in explorerListView.SelectedItems)
                {
                    MazePlayer player = item.Tag as MazePlayer;
                    if (player != null) this.console.Players.Remove(player);
                }

                RefreshPlayerList();
            }
        }
    }
}
