﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Windows.Forms;
using GrapeCity.Competition.Maze.EmbeddedEngine;
using GrapeCity.Competition.Maze.Generators;

namespace GrapeCity.Competition.Maze.UI
{
    internal partial class MazeGameForm : Form, IEngineRecorder
    {
        private MazeGameConsole console;
        private DateTime baseTick;

        private void ResetTicks()
        {
            baseTick = DateTime.Now;
            dataGridView.Rows.Clear();
        }

        private TimeSpan Ticks
        {
            get { return DateTime.Now - baseTick; }
        }

        internal MazeGameForm()
        {
            InitializeComponent();
            console = new MazeGameConsole(this.mazeScreen);

            InitializeRecentMenu();
            InitializeGeneratorList();
        }

        private void MazeGameForm_Load(object sender, EventArgs e)
        {
            console.CreateMaze(21, 21, "Untitled Maze");
            console.AdjustScreenSize(tableLayoutPanel2.Size.Width - 6, tableLayoutPanel2.Size.Height - 6);
            console.Recorder = this;

            UpdateUI();
        }

        private delegate void Write(string name, string action);

        public void WriteRecord(string name, string action)
        {
            if (!checkBoxOutput.Checked)
            {
                return;
            }

            if (this.InvokeRequired)
            {
                this.Invoke(new Write(this.WriteRecord), name, action);
                return;
            }

            try
            {
                dataGridView.Rows.Add(Ticks.TotalSeconds.ToString("0.00"), name, action);
                dataGridView.FirstDisplayedScrollingRowIndex = dataGridView.Rows.Count - 1;
            }
            catch { }
        }

        #region Config
        private void blockColorToolStripSplitButton_ButtonClick(object sender, EventArgs e)
        {
            BlockColorDialog dialog = new BlockColorDialog(this.console.Skin);
            if (dialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                console.Update();
            }
            UpdateUI();
        }

        private void manageAsssemblyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AssemblyLibraryDialog dialog = new AssemblyLibraryDialog(this.console.Assemblies);
            dialog.ShowDialog(this);
            UpdateUI();
        }

        private void explorerToolStripButton_ButtonClick(object sender, EventArgs e)
        {
            ConfigurateExplorerDialog dialog = new ConfigurateExplorerDialog(this.console);
            dialog.ShowDialog(this);
            UpdateUI();
        } 
        #endregion
        #region Update UI
        private void UpdateUI()
        {
            newToolStripMenuItem.Enabled = console.CanCreateMaze;
            openToolStripMenuItem.Enabled = console.CanLoadMaze;
            recentToolStripMenuItem.Enabled = console.CanLoadMaze;
            saveToolStripMenuItem.Enabled = console.CanSaveMaze;
            saveAsToolStripMenuItem.Enabled = console.CanSaveMaze;

            runToolStripButton.Enabled = console.CanRun;
            secondToolStripMenuItem500.Enabled = console.CanRun;
            secondToolStripMenuItem1000.Enabled = console.CanRun;
            noAnimationToolStripMenuItem.Enabled = console.CanRun;

            pauseToolStripButton.Enabled = console.CanPause;
            stopToolStripButton.Enabled = console.CanStop;
            explorerToolStripButton.Enabled = console.CanConfigExplorer;

            if(string.IsNullOrEmpty(console.MapFileName))
            {
                this.Text = "GrapeCity Maze";
            }
            else
            {
                this.Text = string.Format("GrapeCity Maze - {0}", console.MapFileName);
            }
            
            UpdateMazeStats();
            mazeScreen.Refresh();
        }

        private void UpdateMazeStats()
        {
            MazeStatus status = console.Status;
            labelMazeName.Text = string.Format("{0} ({1}x{2})", status.Name, status.Width, status.Height);
            labelComplete.Text = string.Format("{0} ({1}/{2})", status.Complete.ToString("P01"), status.ReachBlock, status.TotalBlock);
            labelComplete.Refresh();
            labelLooks.Text = status.Looks.ToString("N0");
            labelLooks.Refresh();
            labelSteps.Text = status.Steps.ToString("N0");
            labelSteps.Refresh();
            if (status.Steps + status.Blocks == 0)
            {
                labelBlocks.Text = "-";
                labelBlocks.Refresh();
            }
            else
            {
                labelBlocks.Text = string.Format("{0} ({1})", status.Blocks.ToString("N0"), (status.Blocks / (double)(status.Steps + status.Blocks)).ToString("P01"));
                labelBlocks.Refresh();
            }
            
            if(status.Players.Length > 0)
            {
                if (playerListView.Items.Count != status.Players.Length)
                {
                    playerListView.Items.Clear();

                    foreach (MazePlayerStatus player in status.Players)
                    {
                        ListViewItem item = new ListViewItem();
                        item.SubItems.Add("-");
                        item.SubItems.Add("-");
                        item.SubItems.Add("-");
                        playerListView.Items.Add(item);
                    }
                }

                for (int idx = 0; idx < status.Players.Length; idx++)
                {
                    ListViewItem item = playerListView.Items[idx];
                    MazePlayerStatus player = status.Players[idx];
                    item.SubItems[1].Text = player.Name;
                    if(player.X == status.Exit.X && player.Y == status.Exit.Y)
                    {
                        item.SubItems[2].Text = "Completed";
                    }
                    else
                    {
                        item.SubItems[2].Text = string.Format("({0},{1})", player.X, player.Y);
                    }
                    item.SubItems[3].Text = player.Steps.ToString("N0");
                }
            }
            else
            {
                if (playerListView.Items.Count > 0)
                {
                    playerListView.Items.Clear();
                }
            }

            if(status.Players.Length > 1)
            {
                labelMaxSteps.Text = status.Maxstep.ToString("N0");
                labelMaxSteps.Refresh();
                labelMinSteps.Text = status.Minstep.ToString("N0");
                labelMinSteps.Refresh();
                labelAvgSteps.Text = status.Avgstep.ToString("N0");
                labelAvgSteps.Refresh();
            }
            else
            {
                labelMaxSteps.Text = "-";
                labelMinSteps.Text = "-";
                labelAvgSteps.Text = "-";
            }

        }

        private void InitializeGeneratorList()
        {
            Assembly container = typeof(IMazeMapGenerator).Assembly;

            List<MazeGenerator> generators = new List<MazeGenerator>();
            foreach (Type type in container.GetTypes())
            {
                if (type.GetInterface("IMazeMapGenerator") == typeof(IMazeMapGenerator))
                {
                    generators.Add(new MazeGenerator(type));
                }
            }

            generatorToolStripComboBox.Items.AddRange(generators.ToArray());
            if (generatorToolStripComboBox.Items.Count > 0)
                generatorToolStripComboBox.SelectedIndex = 0;
        }

        private void InitializeRecentMenu()
        {
            if (console == null) return;
            if (console.RecentMaps == null || console.RecentMaps.Count == 0)
            {
                recentToolStripMenuItem.Visible = false;
                recentToolStripSeparator.Visible = false;
            }
            else
            {
                recentToolStripMenuItem.Visible = true;
                recentToolStripSeparator.Visible = true;

                //create submenu
                foreach (string recentMap in console.RecentMaps)
                {
                    ToolStripMenuItem item = new ToolStripMenuItem(recentMap);
                    item.Tag = recentMap;
                    item.Click += new EventHandler(recentToolStripMenuItem_Click);
                    recentToolStripMenuItem.DropDownItems.Add(item);
                }
            }
        } 
        #endregion
        #region Create/Open/Save/SaveAs/Recent Actions
        private void recentToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem item = sender as ToolStripMenuItem;
            if (item != null)
            {
                string file = item.Tag as string;
                bool success = console.LoadMaze(file);
                if(!success)
                {
                    MessageBox.Show("The maze map is not exist or damaged, can not open.");
                    console.RecentMaps.Remove(file);
                    UpdateRecentList(null);
                }
                console.AdjustScreenSize(tableLayoutPanel2.Size.Width - 6, tableLayoutPanel2.Size.Height - 6);
            }
            UpdateUI();
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CreateMazeDialog dialog = new CreateMazeDialog();
            if (dialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                console.CreateMaze(dialog.MazeHeight, dialog.MazeWidth, dialog.MazeName, dialog.Entrance, dialog.Exit);
                console.AdjustScreenSize(tableLayoutPanel2.Size.Width - 6, tableLayoutPanel2.Size.Height - 6);
            }
            UpdateUI();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (openFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                console.LoadMaze(openFileDialog.FileName);
                console.AdjustScreenSize(tableLayoutPanel2.Size.Width - 6, tableLayoutPanel2.Size.Height - 6);
                UpdateRecentList(openFileDialog.FileName);
            }
            UpdateUI();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!console.SaveMaze())
            {
                saveFileDialog.FileName = console.Status.Filename;
                if (saveFileDialog.ShowDialog(this) == DialogResult.OK)
                {
                    console.SaveMaze(saveFileDialog.FileName);
                    UpdateRecentList(saveFileDialog.FileName);
                }
            }
            UpdateUI();
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog.FileName = console.Status.Filename;
            if (saveFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                console.SaveMaze(saveFileDialog.FileName);
                UpdateRecentList(saveFileDialog.FileName);
            }
            UpdateUI();
        }

        private void UpdateRecentList(string file)
        {
            if(!string.IsNullOrEmpty(file))
            {
                while(console.RecentMaps.Contains(file))
                {
                    console.RecentMaps.Remove(file);
                }

                console.RecentMaps.Insert(0, file);
            }

            if (console.RecentMaps.Count > 8)
            {
                console.RecentMaps.RemoveAt(8);
            }

            recentToolStripMenuItem.DropDownItems.Clear();
            InitializeRecentMenu();
        }

        #endregion
        #region Close
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        } 
        #endregion
        #region Generate Maze
        private void generateToolStripButton_Click(object sender, EventArgs e)
        {
            MazeGenerator gen = generatorToolStripComboBox.SelectedItem as MazeGenerator;
            if (gen != null)
            {
                int seed;
                if (int.TryParse(seedToolStripTextBox.Text, out seed))
                {
                    console.GenerateMaze(seed, gen.Type);
                }
            }
            UpdateUI();
        }

        private void randomRoolStripButton_Click(object sender, EventArgs e)
        {
            seedToolStripTextBox.Text = (new Random().Next()).ToString();
            generateToolStripButton_Click(sender, e);
        } 
        #endregion
        #region Resize
        private void MazeGameForm_Resize(object sender, EventArgs e)
        {
            try
            {
                if (console != null)
                {
                    console.AdjustScreenSize(tableLayoutPanel2.Size.Width - 6, tableLayoutPanel2.Size.Height - 6);
                }
            }
            catch
            {
            }
        } 
        #endregion
        #region Run/Pause/Stop/Reset Actions
        private void runToolStripButton_ButtonClick(object sender, EventArgs e)
        {
            if (console.CanRun)
            {
                if(console.NeedConfigExplorer)
                {
                    explorerToolStripButton_ButtonClick(sender, e);
                }
                if (!console.NeedConfigExplorer)
                {
                    console.Run(10);
                    ResetTicks();
                    updateTimer.Enabled = true;
                }
                UpdateUI();
            }
        }

        private void secondToolStripMenuItem500_Click(object sender, EventArgs e)
        {
            if (console.CanRun)
            {
                if (console.NeedConfigExplorer)
                {
                    explorerToolStripButton_ButtonClick(sender, e);
                }
                if (!console.NeedConfigExplorer)
                {
                    console.Run(50);
                    ResetTicks();
                    updateTimer.Enabled = true;
                }
                UpdateUI();
            }
        }

        private void secondToolStripMenuItem1000_Click(object sender, EventArgs e)
        {
            if (console.CanRun)
            {
                if (console.NeedConfigExplorer)
                {
                    explorerToolStripButton_ButtonClick(sender, e);
                }
                if (!console.NeedConfigExplorer)
                {
                    console.Run(100);
                    ResetTicks();
                    updateTimer.Enabled = true;
                }
                UpdateUI();
            }
        }

        private void noAnimationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (console.CanRun)
            {
                if (console.NeedConfigExplorer)
                {
                    explorerToolStripButton_ButtonClick(sender, e);
                }
                if (!console.NeedConfigExplorer)
                {
                    console.Run(-1);
                    ResetTicks();
                    updateTimer.Enabled = true;
                }
                UpdateUI();
            }
        }

        private void pauseToolStripButton_Click(object sender, EventArgs e)
        {
            if (console.CanPause)
            {
                console.Pause();
                UpdateUI();
            }
        }

        private void stopToolStripButton_Click(object sender, EventArgs e)
        {
            if (console.CanStop)
            {
                console.Stop();
                updateTimer.Enabled = false;
                UpdateUI();
            }
        }

        #endregion
        #region Themes
        private void defalutThemeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            console.Skin.ResetColor();
            console.Update();
            mazeScreen.Refresh();
        }

        private void darkBlueToolStripMenuItem_Click(object sender, EventArgs e)
        {
            console.Skin.ThemeDarkBlue();
            console.Update();
            mazeScreen.Refresh();
        } 
        #endregion

        private void updateTimer_Tick(object sender, EventArgs e)
        {
            UpdateMazeStats();
            UpdateUI();
        }

        private void MazeGameForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            console.SaveSettings();
            Environment.Exit(0); 
        }

        private void splitter1_SplitterMoved(object sender, SplitterEventArgs e)
        {
            console.AdjustScreenSize(tableLayoutPanel2.Size.Width - 6, tableLayoutPanel2.Size.Height - 6);
        }

        private void MazeGameForm_KeyDown(object sender, KeyEventArgs e)
        {
            try
            {
                if (e.KeyCode == Keys.F5)
                {
                    if (runToolStripButton.Enabled)
                    {
                        runToolStripButton_ButtonClick(sender, e);
                    }
                    else if (pauseToolStripButton.Enabled)
                    {
                        pauseToolStripButton_Click(sender, e);
                    }
                }
                else if (e.KeyCode == Keys.F6)
                {
                    if (stopToolStripButton.Enabled)
                    {
                        stopToolStripButton_Click(sender, e);
                    }
                }
            }
            catch
            {
            }
        }
    }
}
